#include "squidwall.h"

/**
 * Copyright (C) 2005-2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * routines for antivirus checking with clamav
 */

static int clamd_type=0;
#define CLAMAV_NONE   0
#define CLAMAV_LOCAL  1
#define CLAMAV_TCP    2

/**
 * buffers for clamd command and stream connection
 */
static char clamd_iBUF[BUFSIZE]; buffer clamd_in;
static char clamd_oBUF[BUFSIZE]; buffer clamd_out;
static char clams_oBUF[BUFSIZE]; buffer clams_out;

static int64 fd_clamd;  /* command connection to clamd */
static int64 fd_stream; /* our stream to clamav */

static int check_clamav_connect_socket(void);
static int check_clamav_connect_tcp(void);
static int check_clamav_create_stream(void);
static void check_clamav_disconnect(void);
static void check_clamav_getversion(void);


/**
 * check_clamav_connect_tcp - create connection to clamd via tcp
 *
 * @return  0 on success, -1 on failure
 */
static int check_clamav_connect_tcp(void)
{
	debug_s(DEBUG_CHECK_VIRUS, "check_clamav_connect_tcp()");

	fd_clamd = socket_tcp4();
	if (fd_clamd == -1) {
		return -1;
	}

	socket_connect4(fd_clamd, (char*)&clamd_ip, clamd_port);
	if (!socket_connected(fd_clamd)) {
		close(fd_clamd);
		return -1;
	}

	if (!io_fd(fd_clamd)) {
		close(fd_clamd);
		return -1;
	}
	
	return 0;
}


/**
 * check_clamav_connect_socket - create connection to a local clamd socket
 *
 * @return 0 on success, -1 on failure
 */
static int check_clamav_connect_socket(void)
{
	struct sockaddr_un un;

	debug_s(DEBUG_CHECK_VIRUS, "check_clamav_connect_socket()");

	fd_clamd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd_clamd == -1) {
		return -1;
	}

	un.sun_family=AF_UNIX;
	un.sun_path[fmt_strn(un.sun_path, clamav_socket, sizeof(un.sun_path))]=0;
	if (connect(fd_clamd, (struct sockaddr*)&un, sizeof(un)) == -1) {
		close(fd_clamd);
		return -1;
	}

	return 0;
}


/**
 * check_clamav_create_stream - create stream connection
 *
 * @return 0 on success, -1 on failure
 */
static int check_clamav_create_stream(void)
{
	stralloc clamdtext;
	int i;

	debug_s(DEBUG_CHECK_VIRUS, "check_clamav_create_stream()");

	stralloc_init(&clamdtext);

	/* initialize the command connection */
	buffer_init(&clamd_in,   (int(*)())read, fd_clamd,
		clamd_iBUF, sizeof clamd_iBUF);
	buffer_init(&clamd_out, (int(*)())write, fd_clamd,
		clamd_oBUF, sizeof clamd_oBUF);

	io_wantwrite(fd_clamd); io_wait(); io_dontwantwrite(fd_clamd);
	buffer_putsflush(&clamd_out, "STREAM\n");

	io_wantread(fd_clamd); io_wait(); io_dontwantread(fd_clamd);
	if (buffer_getnewline_sa(&clamd_in, &clamdtext) == 0) {
		panic("can't read response of clamd");
		stralloc_free(&clamdtext);
		return -1;
	}
	clamdtext.s[clamdtext.len]=0;

	if (!str_start(clamdtext.s, "PORT ")) {
		panic("bad response from clamd, expected: PORT xxx");
		stralloc_free(&clamdtext);
		return -1;
	}

	i=scan_uint(clamdtext.s+5, &clams_port);
	if (i == 0) {
		panic("can't read port of clamd stream");
		stralloc_free(&clamdtext);
		return -1;
	}

	stralloc_free(&clamdtext);

	fd_stream=socket_tcp4();
	if (fd_stream == -1) {
		panic("socket() for clamd failed");
		return -1;
	}

	/**
	 * open connection to clamd stream socket
	 * - clamd_ip defaults to 127.0.0.1, so it is also valid for socket
	 */
	socket_connect4(fd_stream, (char*)&clamd_ip, clams_port);
	if (!socket_connected(fd_stream)) {
		panic("connect() to clamd failed");
		close(fd_stream);
		return -1;
	}

	/* initialize the stream connection */
	buffer_init(&clams_out, (int(*)())write, fd_stream,
		clams_oBUF, sizeof clams_oBUF);

	if (!io_fd(fd_stream)) {
		panic("io_fd() failed");
		return -1;
	}

	return 0;
}


/**
 * check_clamav_disconnect - disconnect from clamav
 *
 * @return nothing
 */
static void check_clamav_disconnect(void)
{
	debug_s(DEBUG_CHECK_VIRUS, "check_clamav_disconnect()");
	clamd_in.fd=-1;  buffer_close(&clamd_in);
	clamd_out.fd=-1; buffer_close(&clamd_out);
	io_close(fd_clamd);

	return;
}


/**
 * check_clamav_connect - create connection to clamav
 *
 * @return 0 on success, -1 on failure
 */
int check_clamav_connect(void)
{
	int r = -1;

	debug_s(DEBUG_CHECK_VIRUS, "check_clamav_connect()");
	switch (clamd_type) {
	case CLAMAV_NONE:
		r = check_clamav_connect_tcp();
		if (r == 0) {
			log_info("clamav: via tcp");
			clamd_type=CLAMAV_TCP;
			break;
		}
		r = check_clamav_connect_socket();
		if (r == 0) {
			log_info("clamav: via socket file");
			clamd_type=CLAMAV_LOCAL;
			break;
		}
		break;
	case CLAMAV_LOCAL:
		r = check_clamav_connect_socket();
		break;
	case CLAMAV_TCP:
		r = check_clamav_connect_tcp();
		break;
	}

	if (r == -1) {
		log_info("clamav: not available");
		clamd_type=CLAMAV_NONE;
		return r;
	}

	/* create stream, and return result */
	r = check_clamav_create_stream();

	return r;
}


/**
 * check_clamav_put2stream - send one character to clamav stream
 *
 * @param ch  next char, which should be send to clamav
 * @return    nothing
 */
void check_clamav_put2stream(char ch)
{
	if (clamd_type == CLAMAV_NONE)
		return;

	buffer_put(&clams_out, &ch, 1);
	return;
}


/**
 * check_clamav_checkresult - check the stream ...
 * - closes the data stream to clamav
 * - check the result
 *
 * @return 0 => no virus, 1 => virus found
 */
int check_clamav_checkresult(void)
{
	stralloc clamdtext;

	if (clamd_type == CLAMAV_NONE)
		return 0; /* no scanning -> no virus ;) */

	debug_s(DEBUG_CHECK_VIRUS, "check_clamav_check_result()");
	stralloc_init(&clamdtext);

	/* flush and close stream connection */
	buffer_flush(&clams_out);
	clams_out.fd=-1; buffer_close(&clams_out);
	io_close(fd_stream);

	/* read the result on the command connection */
	io_wantread(fd_clamd); io_wait(); io_dontwantread(fd_clamd);
	if (!stralloc_copys(&clamdtext, "")) die_nomem();
	while (buffer_getline_sa(&clamd_in, &clamdtext) != 0);
	clamdtext.s[clamdtext.len]=0;

	check_clamav_disconnect();

	/* "stream: OK\n" && len==11 => no virus! */
	if (clamdtext.len <= 11) {
		stralloc_free(&clamdtext);
		return 0;
	}

	/* "stream: ....\n" - but only 'FOUND\n' means there is a VIRUS! */
	// if (str_equal(clamdtext.s+clamdtext.len-7, "FOUND\n")) {
	// 	stralloc_free(&clamdtext);
	// 	return 0;
	// }

	/* "stream: Worm.Sober.I FOUND\n" -> "Worm.Sober.I" */
	clamdtext.s[clamdtext.len-7]=0; /* remove " FOUND\n" */
	if (!stralloc_copys(&virus, clamdtext.s+8)) die_nomem();
	stralloc_free(&clamdtext);

	/* we have a virus, get current version of clamav */
	check_clamav_getversion();
	return 1;
}


/**
 * check_clamav_getversion - get current version of clamav
 *
 * @return 0 => no virus, 1 => virus found
 */
void check_clamav_getversion(void)
{
	int r = -1;

	debug_s(DEBUG_CHECK_VIRUS, "check_clamav_getversion()");
	switch (clamd_type) {
	case CLAMAV_LOCAL:
		r = check_clamav_connect_socket();
		break;
	case CLAMAV_TCP:
		r = check_clamav_connect_tcp();
		break;
	}

	/* some problem while connecting to clamav ?! */
	if (r == -1) {
		clamd_type=CLAMAV_NONE;
		return;
	}

	/* get curent version of clamav */
	buffer_init(&clamd_in, (int(*)())read, fd_clamd,
		clamd_iBUF, sizeof clamd_iBUF);
	buffer_init(&clamd_out, (int(*)())write, fd_clamd,
		clamd_oBUF, sizeof clamd_oBUF);

	io_wantwrite(fd_clamd); io_wait(); io_dontwantwrite(fd_clamd);
	buffer_putsflush(&clamd_out, "VERSION\n");

	io_wantread(fd_clamd); io_wait(); io_dontwantread(fd_clamd);
	if (buffer_getnewline_sa(&clamd_in, &clamav) == 0) {
		close(fd_clamd);
		panic("can't read response of clamd");
		return;
	}
	clamav.s[clamav.len]=0;

	if (!str_start(clamav.s, "ClamAV ")) {
		close(fd_clamd);
		panic("bad response from clamd, expected: 'ClamAV ' ...");
		return;
	}
	check_clamav_disconnect();

	return;
}
